﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Data;

namespace EditableObjects
{
    /// <summary>Base class for DataObject. Provides interface implementations.</summary>
    public abstract class DataObjectBase : INotifyPropertyChanged, IEditableObject, IDataErrorInfo
    {
        /// <summary>Remembers whether the object is in edit mode, during IEditableObject operations.</summary>
        private bool _isInEditMode = false;
        /// <summary>Contains the backup values during IEditableObject operations.</summary>
        private Dictionary<string, object> _backup;
        /// <summary>Contains the validation errors. Multiple validation errors can be stored per property.</summary>
        private Dictionary<string, List<string>> _errors;

        protected DataObjectBase()
        {
            _backup = new Dictionary<string, object>();
            _errors = new Dictionary<string, List<string>>();
        }

        #region INotifyPropertyChanged

        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged(string propertyName)
        {
            PropertyChangedEventHandler handler = PropertyChanged;

            if (handler != null)
                handler(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion

        #region IEditableObject

        public void BeginEdit()
        {
            this.Trace("IEditableObject.BeginEdit");

            if (!_isInEditMode)
            {
                ClearBackup();
                ExportValues(_backup);
                _isInEditMode = true;
            }
        }

        public void CancelEdit()
        {
            this.Trace("IEditableObject.CancelEdit");

            if (_isInEditMode)
            {
                ImportValues(_backup);
                ClearBackup();
                _isInEditMode = false;
            }
        }

        public void EndEdit()
        {
            this.Trace("IEditableObject.EndEdit");

            if (_isInEditMode)
            {
                ClearBackup();
                _isInEditMode = false;
            }
        }

        /// <summary>Override this method to make a backup of all original values.</summary>
        protected abstract void ExportValues(Dictionary<string, object> backup);
        /// <summary>Override this method to restore all original values.</summary>
        protected abstract void ImportValues(Dictionary<string, object> backup);

        private void ClearBackup()
        {
            _backup.Clear();
        }

        #endregion

        #region IDataErrorInfo

        public string Error
        {
            get
            {
                StringBuilder sb = new StringBuilder();

                foreach (string error in GetErrors())
                    sb.AppendFormat("{0}{1}", sb.Length == 0 ? "" : "\r\n", error);

                return sb.Length == 0 ? null : sb.ToString();
            }
        }

        public string this[string columnName]
        {
            get
            {
                StringBuilder sb = new StringBuilder();

                foreach (string error in GetErrors(columnName))
                    sb.AppendFormat("{0}{1}", sb.Length == 0 ? "" : "\r\n", error);

                return sb.Length == 0 ? null : sb.ToString();
            }
        }

        /// <summary>Gets all validation errors, for all properties.</summary>
        private IEnumerable<string> GetErrors()
        {
            var enumerator = _errors.GetEnumerator();

            while (enumerator.MoveNext())
                foreach (string error in enumerator.Current.Value)
                    yield return error;
        }

        /// <summary>Gets all validation errors for a specific property.</summary>
        private IEnumerable<string> GetErrors(string propertyName)
        {
            if (string.IsNullOrEmpty(propertyName) || !_errors.ContainsKey(propertyName))
                yield break;

            List<string> errors = _errors[propertyName];

            foreach (string error in errors)
                yield return error;
        }

        protected void AddError(string propertyName, string errorMessage)
        {
            List<string> errors;

            if (!_errors.ContainsKey(propertyName))
                _errors.Add(propertyName, new List<string>());
            
            errors = _errors[propertyName];
            errors.Add(errorMessage);
        }

        /// <summary>Removes all errors for all properties.</summary>
        protected void ClearErrors()
        {
            _errors.Clear();
            OnPropertyChanged(string.Empty);
            OnPropertyChanged("Error");
            OnPropertyChanged(Binding.IndexerName);
        }

        /// <summary>Removes all errors for a single property.</summary>
        protected void ClearErrors(string propertyName)
        {
            if (_errors.ContainsKey(propertyName))
            {
                _errors.Remove(propertyName);
            }

            OnPropertyChanged(string.Empty);
            OnPropertyChanged("Error");
            OnPropertyChanged(Binding.IndexerName);
        }

        #endregion
    }
}
